/*  Gtk+ Widget Factory
 *  Copyright (C) 1999-2000  Jeroen Benckhuijsen
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include "gwf_xmlio.h"
#include "gwf_project.h"
#include "gwf_tools.h"
#include "gwf_error.h"
#include <gtk/gtk.h>
#include <tree.h>
#include <parser.h>

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

enum
{
   GWF_PROJECT,
   GWF_MAIN_VERSION,
   GWF_SECTION,
   GWF_SECTION_NAME,
   GWF_SECTION_VERSION,
   GWF_SECTION_HREF,

   LAST_TAG
};

static gchar		*xml_tags[LAST_TAG] =	{
						"PROJECT",
						"VERSION",
						"SECTION",
						"NAME",
						"VERSION",
						"HREF"
						};

static gchar		*prefix = "GWF";

static gboolean		gwf_xmlio_free_loaders		(gpointer key,gpointer value,gpointer user_data);
static gboolean		gwf_xmlio_free_savers		(gpointer key,gpointer value,gpointer user_data);
static gint		gwf_xmlio_parse_section		(xmlNodePtr cur,xmlNsPtr ns,GwfXmlIO *xml_io,GwfProject *project,GwfError *error);
static gint		gwf_xmlio_check_version		(gchar *loadver,gchar *filever);
static void		gwf_xmlio_section_save_foreach	(gpointer key,gpointer value,gpointer user_data);
static void		free_xml_loader			(GwfXmlLoader *loader);
static void             free_xml_saver			(GwfXmlSaver *saver);


GwfXmlIO*
gwf_xmlio_new (void)
{
  GwfXmlIO		*tmp;

  tmp = g_malloc(sizeof(GwfXmlIO));

  tmp->buffer = NULL;
  tmp->doc = NULL;
  tmp->loaders = g_hash_table_new(g_str_hash,g_str_equal);
  tmp->savers = g_hash_table_new(g_str_hash,g_str_equal);

  return tmp;
}

void
gwf_xmlio_free (GwfXmlIO *xml_io)
{
  g_return_if_fail(xml_io != NULL);

  g_free(xml_io->buffer);
  xmlFreeDoc(xml_io->doc);

  g_hash_table_foreach_remove(xml_io->loaders,gwf_xmlio_free_loaders,NULL);
  g_hash_table_foreach_remove(xml_io->savers,gwf_xmlio_free_savers,NULL);
  g_hash_table_destroy(xml_io->loaders);
  g_hash_table_destroy(xml_io->savers);
  g_free(xml_io);
}

gint
gwf_xmlio_add_loader (GwfXmlIO *xml_io,gchar *name,gchar *version,gchar *href,GwfXmlLoadFunc func)
{
  GwfXmlLoader		*tmp;

  g_return_val_if_fail(xml_io != NULL,0);
  g_return_val_if_fail(name != NULL,0);
  g_return_val_if_fail(version != NULL,0);
  g_return_val_if_fail(href != NULL,0);
  g_return_val_if_fail(func != NULL,0);

  if (g_hash_table_lookup(xml_io->loaders,(gconstpointer)name))
    return 0;

  tmp = g_malloc(sizeof(GwfXmlLoader));

  tmp->name = g_strdup(name);
  tmp->version = g_strdup(version);
  tmp->href = g_strdup(href);
  tmp->func = func;
  g_hash_table_insert(xml_io->loaders,(gpointer)tmp->name,(gpointer)tmp);

  return 1;
}

void
gwf_xmlio_del_loader (GwfXmlIO *xml_io,gchar *name)
{
  GwfXmlLoader		*tmp;

  g_return_if_fail(xml_io != NULL);
  g_return_if_fail(name != NULL);

  tmp = g_hash_table_lookup(xml_io->loaders,(gconstpointer)name);
  if (!tmp)
    return;

  g_hash_table_remove(xml_io->loaders,(gconstpointer)name);
  free_xml_loader(tmp);
}


gint
gwf_xmlio_add_saver (GwfXmlIO *xml_io,gchar *name,gchar *version,gchar *href,GwfXmlSaveFunc func)
{
  GwfXmlSaver		*tmp;

  g_return_val_if_fail(xml_io != NULL,0);
  g_return_val_if_fail(name != NULL,0);
  g_return_val_if_fail(version != NULL,0);
  g_return_val_if_fail(href != NULL,0);
  g_return_val_if_fail(func != NULL,0);

  if (g_hash_table_lookup(xml_io->savers,(gconstpointer)name))
    return 0;

  tmp = g_malloc(sizeof(GwfXmlSaver));

  tmp->name = g_strdup(name);
  tmp->version = g_strdup(version);
  tmp->href = g_strdup(href);
  tmp->func = func;
  g_hash_table_insert(xml_io->savers,(gpointer)tmp->name,(gpointer)tmp);

  return 1;
}

void
gwf_xmlio_del_saver (GwfXmlIO *xml_io,gchar *name)
{
  GwfXmlSaver		*tmp;

  g_return_if_fail(xml_io != NULL);
  g_return_if_fail(name != NULL);

  tmp = g_hash_table_lookup(xml_io->savers,(gconstpointer)name);
  if (!tmp)
    return;

  g_hash_table_remove(xml_io->savers,(gconstpointer)name);
  free_xml_saver(tmp);
}

void
gwf_xmlio_set_buffer (GwfXmlIO *xml_io,guchar *buffer)
{
  g_return_if_fail(xml_io != NULL);

  g_free(xml_io->buffer);
  xml_io->buffer = buffer;
}

guchar*
gwf_xmlio_get_buffer (GwfXmlIO *xml_io)
{
  g_return_val_if_fail(xml_io != NULL,NULL);
  return xml_io->buffer;
}

static void
free_xml_loader(GwfXmlLoader *loader)
{
  g_free(loader->name);
  g_free(loader->version);
  g_free(loader->href);
  g_free(loader);
}

static void
free_xml_saver(GwfXmlSaver *saver)
{
  g_free(saver->name);
  g_free(saver->version);
  g_free(saver->href);
  g_free(saver);
}

static gboolean
gwf_xmlio_free_loaders(gpointer key,gpointer value,gpointer user_data)
{
  free_xml_loader((GwfXmlLoader*)value);
  return TRUE;
}

static gboolean
gwf_xmlio_free_savers(gpointer key,gpointer value,gpointer user_data)
{
  free_xml_saver((GwfXmlSaver*)value);
  return TRUE;
}



/******************/
/* Load functions */
/******************/

GwfProject*
gwf_xmlio_load (GwfXmlIO *xml_io,GwfProjlist *projlist,GwfError *error)
{
  xmlNodePtr		cur;
  GwfProject		*project;
  gchar			*version;
  gint			ok = 1;
  xmlNsPtr 		ns;
  gchar			*name;

  g_return_val_if_fail(xml_io != NULL,NULL);
  g_return_val_if_fail(xml_io->buffer != NULL,NULL);

  xml_io->doc = xmlParseDoc(xml_io->buffer);
  if (!xml_io->doc)
    return NULL;

  cur = xml_io->doc->root;
  if (!cur || (strcmp(cur->name,xml_tags[GWF_PROJECT]) != 0))
  {
    gwf_error_raise_error(error,"XmlIO : Invalid file");
    xmlFreeDoc(xml_io->doc);
    return NULL;
  }

  version = xmlGetProp(cur,xml_tags[GWF_MAIN_VERSION]);
  if (!gwf_xmlio_check_version(version,XML_FILE_VERSION))
  {
    gwf_error_raise_error(error,"XmlIO : Program version");
    xmlFreeDoc(xml_io->doc);
    g_free(version);
    return NULL;
  }
  g_free(version);


  ns = xmlSearchNsByHref(xml_io->doc, cur, XML_HREF);
  if (ns == NULL)
  {
    xmlFreeDoc(xml_io->doc);
    return NULL;
  }

  project = gwf_project_new();
  name = gwf_projlist_unique_widname(projlist);
  gwf_project_set_widname(project,name);
  g_free(name);

  if (!gwf_projlist_add_project(projlist,project))
  {
    gwf_error_raise_error(error,"General : Widget exists");
    gtk_object_destroy(GTK_OBJECT(project));
    return NULL;
  }

  cur = cur->childs;
  while (cur  && ok)
  {
    if (strcmp(cur->name,xml_tags[GWF_SECTION]) == 0)
      ok = gwf_xmlio_parse_section(cur,ns,xml_io,project,error);
    else
    {
      gwf_error_raise_error(error,"XmlIO : Unknown tag",cur->name);
      ok = 0;
    }

    cur = cur->next;
  }

  if (!gwf_project_get_widname(project))
    gwf_error_raise_error(error,"XmlIO : No name");

  if (!ok || !gwf_project_get_widname(project))
  {
    gtk_object_destroy(GTK_OBJECT(project));
    xmlFreeDoc(xml_io->doc);
    return NULL;
  }

  return project;
}

static gint
gwf_xmlio_check_version(gchar *loadver,gchar *filever)
{
  gint			loadmajor,loadminor,loadmicro;
  gint			filemajor,fileminor,filemicro;

  g_return_val_if_fail(loadver != NULL,0);
  g_return_val_if_fail(filever != NULL,0);

  sscanf(filever,"%d.%d.%d",&filemajor,&fileminor,&filemicro);
  sscanf(loadver,"%d.%d.%d",&loadmajor,&loadminor,&loadmicro);

  if (loadmajor > filemajor)
    return 0;

  if ((loadmajor = filemajor) && (loadminor > fileminor))
    return 0;

  return 1;
}

static gint
gwf_xmlio_parse_section(xmlNodePtr cur,xmlNsPtr ns,GwfXmlIO *xml_io,GwfProject *project,GwfError *error)
{
  gchar		*name;
  gchar		*version;
  gchar		*href;
  GwfXmlLoader	*loader;

  g_return_val_if_fail(cur != NULL,0);
  g_return_val_if_fail(project != NULL,0);

  name = xmlGetProp(cur,xml_tags[GWF_SECTION_NAME]);
  loader = g_hash_table_lookup(xml_io->loaders,(gpointer)name);
  g_free(name);

  href = xmlGetProp(cur,xml_tags[GWF_SECTION_HREF]);
  if (strcmp(href,loader->href) != 0)
  {
    g_free(href);
    gwf_error_raise_error(error,"XmlIO : Invalid Href",loader->href);
    return 0;
  }

  if (!loader)
  {
    gwf_error_raise_error(error,"XmlIO : Unknown loader",href);
    g_free(href);
    return 0;
  }

  version = xmlGetProp(cur,xml_tags[GWF_SECTION_VERSION]);
  if (!gwf_xmlio_check_version(version,loader->version))
  {
    g_free(version);
    gwf_error_raise_error(error,"XmlIO : Wrong version",href);
    g_free(href);
    return 0;
  }
  g_free(version);

  g_free(href);
  return (cur->childs) ? loader->func(cur->childs,ns,xml_io,project,error) : 1;
}


/******************/
/* Save Functions */
/******************/

gint
gwf_xmlio_save (GwfXmlIO *xml_io,GwfProject *project)
{
  gint			size;
  XmlSaveData		save_data;
  xmlNsPtr		ns;

  g_return_val_if_fail(xml_io != NULL,0);
  g_return_val_if_fail(project != NULL,0);

  g_free(xml_io->buffer);
  if (xml_io->doc)
    xmlFreeDoc(xml_io->doc);

  xml_io->doc = xmlNewDoc("1.0");
  xml_io->doc->root = xmlNewDocNode(xml_io->doc,NULL,xml_tags[GWF_PROJECT],NULL);
  ns = xmlNewNs(xml_io->doc->root,XML_HREF,prefix);
  if (!ns)
  {
     xmlFreeDoc(xml_io->doc);
     return 0;
  }
  xmlSetNs(xml_io->doc->root,ns);
  xmlSetProp(xml_io->doc->root,xml_tags[GWF_MAIN_VERSION],XML_FILE_VERSION);

  save_data.ns = ns;
  save_data.project = project;
  save_data.parent = xml_io->doc->root;
  save_data.result = 1;

  g_hash_table_foreach(xml_io->savers,gwf_xmlio_section_save_foreach,(gpointer)&save_data);

  if (!save_data.result)
  {
     xmlFreeDoc(xml_io->doc);
     return 0;
  }

  xmlDocDumpMemory(xml_io->doc,&(xml_io->buffer),&size);

  return 1;
}

static void
gwf_xmlio_section_save_foreach(gpointer key,gpointer value,gpointer user_data)
{
  XmlSaveData		*parent_data = user_data;
  XmlSaveData		data;
  GwfXmlSaver		*saver = value;
  xmlNodePtr		sect;

  sect = xmlNewChild(parent_data->parent,parent_data->ns,xml_tags[GWF_SECTION],NULL);
  xmlSetProp(sect,xml_tags[GWF_SECTION_NAME],saver->name);
  xmlSetProp(sect,xml_tags[GWF_SECTION_VERSION],saver->version);
  xmlSetProp(sect,xml_tags[GWF_SECTION_HREF],saver->href);

  data.ns = parent_data->ns;
  data.project = parent_data->project;
  data.parent = sect;

  data.result = saver->func(data.parent,data.ns,data.project);
  if (!data.result)
    parent_data->result = 0;
}
